"""21cmEMU: An emulator of 21cmFAST summaries."""

__version__ = "1.0.0"
from .emulator import Emulator
from .get_emulator import add_emulator_data
from .get_emulator import download_emu_data
from .get_emulator import move_emulator_data
